"""Use linear SVM for iris classification."""
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from sklearn.datasets import load_iris
from sklearn.metrics import ConfusionMatrixDisplay
from sklearn.model_selection import train_test_split
from sklearn.svm import LinearSVC


def train_test_svc(
    xtrain: np.ndarray, xtest: np.ndarray, ytrain: np.ndarray, ytest: np.ndarray
) -> Tuple[LinearSVC, float]:
    """Train a linear SVM classifier and compute the accuracy on the test set.

    Define a linear SVM classifier. Train it and compute the accuracy on the test set.
    Express the accuracy as a percentage and round to one decimal place.

    Args:
        xtrain (np.ndarray): The training data.
        xtest (np.ndarray): The test data.
        ytrain (np.ndarray): The training labels.
        ytest (np.ndarray): The test labels.

    Returns:
        Tuple[LinearSVC, float]: The trained linear SVM classifier and the accuracy of the model on the test set.
    """
    # 4.1. create and train linear SVM
    clf = LinearSVC(dual=False)
    clf.fit(xtrain, ytrain)

    # 4.2. predict on test set
    ypred = clf.predict(xtest)
    # 4.2. calculate accuracy
    acc = (ytest == ypred).mean() * 100
    acc = acc.round(1)
    return clf, acc


if __name__ == "__main__":
    # 1. load the Iris dataset
    iris_data = load_iris()

    # 2. get access to data, labels and class names
    x = iris_data.data
    y = iris_data.target
    target_names = iris_data.target_names

    # 3. split data into training and test sets
    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.25, random_state=29
    )

    # 4. use function train_test_svc to train a linear SVM model and calculate accuracy on test set
    model, accuracy = train_test_svc(x_train, x_test, y_train, y_test)
    # 5. print accuracy
    print("The model has an accuracy of {}%".format(accuracy))

    # 6. plot confusion matrix
    ConfusionMatrixDisplay.from_estimator(
        model,
        x_test,
        y_test,
        display_labels=target_names,
        xticks_rotation="vertical",
    )

    plt.tight_layout()
    plt.savefig("./figures/linear_svm.png")
    plt.show()
